/***************************************************************************
                          cachablepixmaparray.cpp  -  description
                             -------------------
    begin                : Fri Oct 25 2002
    copyright            : (C) 2002, 2004 by Thomas Friedrichsmeier
    email                : tfry@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cachablepixmaparray.h"

#include <klocale.h>
#include <qregexp.h>
#include <qbitmap.h>

#include "cdp.h"

QMap<QString, CachablePixmapArray *> CachablePixmapArray::cache_map;

CachablePixmapArray::CachablePixmapArray() {
	references = 1;   // namely itself
}

CachablePixmapArray::~CachablePixmapArray(){
	// remove entry from cache
	CacheMap::Iterator it;
	QString to_delete = "";
	if (cache_map.count ()) {
		for (it = cache_map.begin(); it != cache_map.end(); it++) {
			if (it.data () == this) {
				to_delete = it.key ();
				// place iterator next to end
				it = cache_map.end ();
				it--;
			}
		}
	}
		// deleting the entry right within
		// the for-loop messes up the iterators
		// operator++. So we have to delay to here.
	if (to_delete != "") cache_map.remove (to_delete);
}

CachablePixmapArray* CachablePixmapArray::create (const QString & datafilenamepattern, int fc, int def_width, int def_height, bool preserve_alpha) {
	if (cache_map.contains (datafilenamepattern)) {
		cdp->debug_msg (i18n ("Found animation '%1' in cache.").arg(datafilenamepattern), Info);
		CachablePixmapArray *cached = cache_map[datafilenamepattern];
		if (cached->count () < static_cast<unsigned int> (fc)) {
// Qt-bug (3.0.4): can't reload an existing QCanvasPixmapArray! ->crash
			cdp->debug_msg (i18n ("Should rebuild cached animation with more frames, which is not currently possible"), Warn);
//			cached->build (datafilenamepattern, fc);
		}
		cached->references++;
		return (cached);
	} else {
		CachablePixmapArray *new_pixmap = new CachablePixmapArray ();
		if (!new_pixmap->build (datafilenamepattern, fc, preserve_alpha)) {
			cdp->debug_msg (i18n ("PixmapArray '%1' could not be read. Creating a single blank frame instead.").arg(datafilenamepattern), Critical);
// I'd like to delete the messed up array here, but Qt (3.0.4) would crash on that!
			new_pixmap = new CachablePixmapArray ();
			new_pixmap->buildEmpty (def_width, def_height);
		}
		cache_map.insert (datafilenamepattern, new_pixmap);
		return new_pixmap;
	}
}

void CachablePixmapArray::discard (CachablePixmapArray *array) {
	if (--array->references <= 0) {
		if (array->isValid ()) {
			delete array;
		}
	}
}

CachablePixmapArray* CachablePixmapArray::copy (CachablePixmapArray *array) {
	array->references++;
	return array;
}

void CachablePixmapArray::resetCache () {
	CachablePixmapArray::cache_map.clear ();
}

// very ugly hack needed to strip alpha-channels on Qt < 3.0.4 part 1
// not portable!
#if QT_VERSION < 0x030004
#include <qpixmap.h>

class StrippablePixmap : public QPixmap {
public:
	void stripAlphaChannel () {
		if (data->alphapm) {
			delete data->alphapm;
			data->alphapm = 0;
		}
	};
};
#endif

bool CachablePixmapArray::build (const QString & datafilenamepattern, int fc, bool preserve_alpha) {
	if (!readPixmaps (datafilenamepattern, fc)) return false;

	if (preserve_alpha) return true;

	// preserve_alpha not set. Reduce alpha-channels to 1-bit mask for better performance
	for (unsigned int i = 0; i < count (); i++) {
// very ugly hack needed to strip alpha-channels on Qt < 3.0.4 part 2
#if QT_VERSION < 0x030004
		static_cast<StrippablePixmap *> (static_cast<QPixmap *> ((image (i))))->stripAlphaChannel ();
#else
 #if QT_VERSION >= 0x030200
		if (image (i)->hasAlphaChannel ()) {
 #endif
			QBitmap maskcopy = QBitmap (*image (i)->mask ());
			image (i)->setMask (maskcopy);
 #if QT_VERSION >= 0x030200
		}
 #endif
#endif
	}
	return true;
}

void CachablePixmapArray::buildEmpty (int width, int height) {
	QCanvasPixmap * dummy = cdp->createQCPixmap (bad_string_value, width, height, Suppress);
	setImage (0, dummy);
}
