/***************************************************************************
                          event.cpp  -  description
                             -------------------
    begin                : Fri Nov 23 2001
    copyright            : (C) 2001, 2002 by Thomas Friedrichsmeier
    email                : Thomas.Friedrichsmeier@ruhr-uni-bochum.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "defines.h"

#include "event.h"

#include <qdom.h>
#include <klocale.h>

#include "level.h"
#include "cdp.h"
#include "state.h"
#include "conditional.h"
#include "qstrlist.h"
#include "soundwrapper.h"

Event::Event(const QDomElement *description) {
	read_event (description);
	active = false;
}

Event::~Event(){
	if (num_changes) {delete [] changes;}
	if (has_conditional) {delete conditional;}
	if (num_child_events) {delete [] child_events;}
}

/** Reads in the description for this event */
void Event::read_event (const QDomElement *description){

	QDomNodeList list;
	QDomElement e;

	ident = cdp->get_string_attribute ("ident", *description, bad_string_value, Serious);

	list = cdp->get_node_list("change", *description, 0, -1, 1, Warn);
	num_changes = 0;

	if (list.length ()) {
		changes = new Changes [list.length()];

		while (num_changes < list.length ()) {		// get the list of states to be changed
			e = list.item(num_changes).toElement ();
	
			QString dummy = cdp->get_string_attribute ("state", e, bad_string_value, Serious);
			changes[num_changes].state = level->resolve_state (dummy);
			dummy = cdp->get_multi_choice_attribute ("to", e, "toggle;on;off", "toggle", Serious);
			if (dummy == "toggle") {
				changes[num_changes].change_type = 0;
			} else if (dummy == "on") {
				changes[num_changes].change_type = 1;
			} else {
				changes[num_changes].change_type = -1;
			}
			num_changes++;
		}
	}

	list = cdp->get_node_list("child_event", *description, 0, -1, 1, Warn);
	num_child_events = 0;

	while (num_child_events < list.length ()) { 		// get the list of child-events
		e = list.item (num_child_events).toElement();

		child_event_idents.append (cdp->get_string_attribute ("ident", e, bad_string_value, Serious));
		num_child_events++;
	}

	list = cdp->get_node_list("condition", *description, 0, 1, 1, Warn);	// get conditional, if any
	if (list.length ()) {
		has_conditional = true;
		e = list.item (0).toElement ();
		conditional = new Conditional (e);
 	} else {
		has_conditional = false;
	}

	sound.create (*description, DefaultSingleSound, Info);		// read sound for this event
}

/** Triggers this event */
void Event::trigger(){
	if (active) {
		cdp->debug_msg (i18n ("Caught recursion in event '%1'!").arg(ident), Serious);
	} else {
		if (has_conditional) {
			if (!conditional->fulfilled ()) {
				return;
			}
		}
		active = true;
		sw->start_sound (sound.id[0]);
		unsigned int i;
		for (i=0; i < num_changes; i++) {
			if (!changes[i].change_type) {
				changes[i].state->toggle ();
			} else {
				changes[i].state->set (changes[i].change_type > 0);
			}
		}
		for (i=0; i < num_child_events; i++) {
			child_events[i]->trigger ();
		}
		active = false;
	}
}

/** Tells the event, that it can now resolve it's child events. Called by Level, when
all event-declarations have been read. */
void Event::resolve_now (){

	if ((num_child_events = child_event_idents.count ())) {
		child_events = new Event * [num_child_events];
	
		for (unsigned int i = 0; i < num_child_events; i++) {
			QString dummy = child_event_idents.at(i);
			child_events[i] = level->resolve_event (dummy);
		}	
	
		child_event_idents.clear ();		// This list is no longer needed
	}

}
