/***************************************************************************
                          taxipilot.cpp  -  description
                             -------------------
    begin                : Don Mai 17 20:03:43 CEST 2001
    copyright            : (C) 2001, 2002, 2004, 2005, 2007 by Thomas Friedrichsmeier
    email                : Thomas.Friedrichsmeier@ruhr-uni-bochum.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "defines.h"

#include <klocale.h>
#include <kmessagebox.h>
#include <kfiledialog.h>
#include <kapplication.h>

#include <qlayout.h>
#include <qdom.h>
#include <qrect.h>
#include <qcolor.h>
#include <qfont.h>
//#include <qpalette.h>

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

#include "taxipilot.h"

#include "taxi.h"
#include "level.h"
#include "splashscreen.h"
#include "passenger.h"
#include "cdp.h"
#include "taxipilotapp.h"
#include "movingobject.h"
#include "teleporter.h"
#include "centerofgravity.h"
#include "platform.h"
#include "message.h"
#include "timemanager.h"
#include "hotarea.h"
#include "tpbase.h"
#include "soundwrapper.h"
#include "miscwidgets.h"
#include "missionselector.h"
#include "credits.h"
#include "savegames.h"
#include "moneylcd.h"
#include "highscores.h"

Taxipilot::Taxipilot (TaxipilotApp * parent, const char *name) : QWidget (parent, name) {
	// create global objects (or simply init the pointers from TPBase)
	app = parent;
	tp = this;
	canvas = new QCanvas ();
	view = new QCanvasView(canvas, this);
	cdp = new Cdp();
	tm = new TimeManager (100, 10);
	message_handler = new Message ();
	sw = new SoundWrapper ();
	saves = new SaveGames ();
	highscores = new Highscores ();
	taxi = 0;
	screen = 0;
	current_level = -1;
	scrolling = false;

	current_key_lr = current_key_ud = 0;
	toggle_gear = false;
	game_running = false;
	game_paused = true;
	start_screen_active = false;
	passengers_created = false;
	passenger_exists = false;
	base_timer = passenger_timer = canvas_timer = -1;
	passengers_on_start_screen.setAutoDelete (true);
	taxis_on_start_screen.setAutoDelete (true);
	splashes.setAutoDelete (true);

	srand((unsigned int) time((time_t *) NULL));		// randomize

	QHBoxLayout *centralhbox = new QHBoxLayout(this);
	QVBoxLayout *mainvbox = new QVBoxLayout ();
	QVBoxLayout *rightvbox = new QVBoxLayout ();
	QHBoxLayout *bottomhbox = new QHBoxLayout ();

	view->setHScrollBarMode(QScrollView::AlwaysOff);
	view->setVScrollBarMode(QScrollView::AlwaysOff);

	fader = new QCanvasRectangle(canvas);
	fader->setPen(NoPen);
	fader->setBrush(QBrush::QBrush(Dense4Pattern));
	fader->setZ(FADER_Z);		// second topmost

	mainvbox->addWidget(view);

	QFont labelFont(LABEL_FONT, LABEL_FONT_SIZE);

	QLabel *label;
	label = new QLabel(i18n("Ships"), this);
	label->setFont(labelFont);
	rightvbox->addWidget(label);

	shipsLCD = new QLCDNumber (7, this);
	shipsLCD->setSegmentStyle (QLCDNumber::Flat);
	rightvbox->addWidget(shipsLCD);

	label = new QLabel(i18n("Funds"), this);
	label->setFont(labelFont);
	rightvbox->addWidget(label);

	scoreLCD = new MoneyLCD (this);
	rightvbox->addWidget(scoreLCD);

	label = new QLabel(i18n("Fuel"), this);
	label->setFont(labelFont);
	rightvbox->addWidget(label);

	fuelDial = new EQDial(0, 100, 20, 1, this);
	fuelDial->setNotchTarget(10);
	fuelDial->setLimits(10, -1);
	fuelDial->setNotchesVisible(true);
	fuelDial->setTracking(false);
	fuelDial->setFocusPolicy(QWidget::NoFocus);
	rightvbox->addWidget(fuelDial);

	label = new QLabel(i18n("Speed"), this);
	label->setFont(labelFont);
	rightvbox->addWidget(label);

	speedDial = new EQDial(0, 100, 20, 30, this);
	speedDial->setNotchesVisible(true);
	speedDial->setTracking(false);
	speedDial->setFocusPolicy(QWidget::NoFocus);
	rightvbox->addWidget(speedDial);

	label = new QLabel(i18n("Fare"), this);
	label->setFont(labelFont);
	rightvbox->addWidget(label);

	fareLCD = new MoneyLCD (this);
	rightvbox->addWidget(fareLCD);

	label = new QLabel(i18n("Tip"), this);
	label->setFont(labelFont);
	rightvbox->addWidget(label);

	tipLCD = new MoneyLCD (this);
	rightvbox->addWidget(tipLCD);
	
	label = new QLabel(i18n("Passengers"), this);
	label->setFont(labelFont);
	rightvbox->addWidget(label);
	
	deliveries_LCD = new QLCDNumber (7, this);
	deliveries_LCD->setSegmentStyle (QLCDNumber::Flat);
	rightvbox->addWidget (deliveries_LCD);

	label = new QLabel(i18n("Status:"), this);
	label->setFont(labelFont);
	bottomhbox->addWidget(label);

	statusLabel = new QLabel (this);
	statusLabel->setFont(labelFont);
	statusLabel->setAlignment(AlignRight);
	bottomhbox->addWidget(statusLabel);

	bottomhbox->addSpacing(20);

	label = new QLabel(i18n("Destination:"), this);
	label->setFont(labelFont);
	bottomhbox->addWidget(label);

	destinationLabel = new NotifyingLabel (this);
	destinationLabel->setFont(labelFont);
	destinationLabel->setAlignment(AlignRight);
	bottomhbox->addWidget(destinationLabel);

	mainvbox->addLayout(bottomhbox);

	rightvbox->addSpacing(destinationLabel->height());

	centralhbox->addLayout(mainvbox);
	centralhbox->addLayout(rightvbox);

	init_timer = tm->startTimer (this, 10);		// this timer will get active after a.exec () in main and then fire initialization.
}

Taxipilot::~Taxipilot()
{
	deletePassengers ();

	levels.clear ();
	splashes.clear ();

	if (start_screen_active) {
		delete_start_screen ();
	}

	delete fader;
	delete message_handler;

	deleteTaxi ();
	deleteScreen ();

	delete tm;
	delete cdp;
	delete sw;
	delete saves;
	delete highscores;

	delete view;
	delete canvas;
}

/** This function really would be part of the constructor. It initializes Taxipilot by reading
the default mission, displaying the start-screen, and setting a few variables.
The thing is though, that it may also set the KAction gameNew to enabled or
disabled, which rests on this Action having been created in TaxiPIlotApp, which
in turn depends on having an instance of Taxipilot.
If you can think of something more elegant, please don't hesitate to tell me. */
void Taxipilot::init()
{
	configChanged ();		// this updates what we store about the config (and inits some things)

	setFocusPolicy(QWidget::StrongFocus);	// get Focus
	setFocus();

	if (cdp->selectMissionOnStartup () || (cdp->getMission () == "") || cdp->firstStartThisVersion ()) {
		MissionSelector *sel = new MissionSelector (cdp->getMissionList ().empty () || cdp->firstStartThisVersion (), MissionSelector::Mission, true);
		app->showTipOfDay (cdp->firstStartThisVersion (), sel);
		sel->exec ();
		delete sel;
	} else {
		app->showTipOfDay (false, this);
	}

	app->show ();
	readMission(cdp->prefix_check(cdp->getMission (), Serious));
	initMission ();
}

bool Taxipilot::onActivePlatform()
{
	return (taxi->landed && (on_platform == active_platform));
}

/* This function is to handle virtually everything of significance */
void Taxipilot::timerTick(int id)
{

	if (!game_paused) {	// do nothing if paused.
		if (id == canvas_timer) {	// scroll if necessary and update canvas
			doScroll();
			canvas->update ();
		} else if (id == base_timer) {	// base timer

			// whatever may happen, do the chores ( gravity only if not landed )
			do_chores();

			int might_be_landed = 0;		// number of potential platforms, taxi is in contact with
									// of course, > 1 means crashed

			// check for moving objects here
			for (unsigned int i = 0; i < level->num_objects; i++) {
				if (level->objects[i]->harmful()) {
					if (taxi->collidesWith(level->objects[i])) {
						if (level->objects[i]->platform) {
							might_be_landed++;
						} else {
							crash(CRASH_BACKGR);
							return;
						}
					} else {
						if (taxi->gear_avail && taxi->gear_state) {
							if (taxi->gear->collidesWith(level->objects[i])) {
								if (level->objects[i]->platform) {
									might_be_landed++;
								} else {
									crash(CRASH_BACKGR);
									return;
								}
							}
						}
					}
				}
			}

			// check for hot-areas here
			QRect dummy = taxi->boundingRect ();
			for (unsigned int i=0; i < level->num_areas; i++) {
				level->areas[i]->check (dummy);
			}

			if (taxi->ready) {	// means: not teleporting already

				for (unsigned int i = 0; i < level->num_teleporters; i++) {
					if (level->teleporters[i]->check_teleportation (taxi, &score)) {
						scoreLCD->setValue (score);
						i = level->num_teleporters; // skip further checks
					}
				}

			}

			if (!taxi->landed) {
				if (passenger_exists && (!passenger_onboard)) {	// checking when passenger undefined crashes the program!
					if (taxi->collidesWith(passenger)) {
						crash(CRASH_PASSENGER);
						return;	// skip everything else, it might crash if taxi got deleted
					}
					if (taxi->gear_state && taxi->gear_avail) {
						if (taxi->gear->collidesWith(passenger)) {
							crash(CRASH_PASSENGER);
							return;
						}
					}
					if (current_key_lr) {	// only check for collision with flying anim, when accelerating
						if (taxi->flying_lr->collidesWith(passenger)) {
							crash(CRASH_PASSENGER);
							return;	// skip everything else, it might crash if taxi got deleted
						}
					}
					if (current_key_ud) {
						if (taxi->flying_ud->collidesWith(passenger)) {
							crash(CRASH_PASSENGER);
							return;	// skip everything else, it might crash if taxi got deleted
						}
					}
				}
			}

			if (!taxi->landed) {
				if (taxi->collidesWith(level)) {
					might_be_landed++;
				} else if (taxi->gear_avail) {
					if (taxi->gear_state && taxi->gear->collidesWith(level)) {
						might_be_landed++;
					}
				}
			}

			if (might_be_landed) {
				if (might_be_landed > 1) {		// in contact with more than one potential platform? -> crash!
					crash(CRASH_BACKGR);
					return;		// omit gravity
				} else {
					if (taxi->gear_state == 2) {	// otherwise this is definitely a crash
						if ((on_platform = check_landed()) >= 0) {
							landing_func();		// do platform stuff
						} else {
							crash(CRASH_BACKGR);
							return;
						}
					} else {
						crash(CRASH_BACKGR);
						return;
        				}
				}
			}

			if (!taxi->landed) {
				// nothing interesting happend (and we are flying)
				do_gravity();
			} else {			// hey, we're still landed. anything special to do?
				if (fueling) {
					score -= level->platform[on_platform]->fuel_price * taxi->fuel_up (level->platform[on_platform]->fuel_step);
					fuelDial->setValue (taxi->fuel_state());
					scoreLCD->setValue (score);
				}
			}

		} else if (id == passenger_timer) {	// create new passenger

			int rand_num;
			do {
				rand_num = rand() % (level->num_platforms);
			} while ((!level->platform[rand_num]->accepts_passenger));
			active_platform = rand_num;

			passenger = selectPassenger (levels[current_level].passenger_ids);

			passenger->initialize (level->platform[active_platform]->current_plat_base_x (), level->platform[active_platform]->current_plat_y (), level->platform[active_platform]);
			passenger_exists = true;

			tm->killTimer(passenger_timer);
			passenger_timer = -1;

			// everyting else is taken care of by passenger_ready ()

		}
	} else {					// stuff to do only when game is paused
		if (id == canvas_timer) {	// advance credits and update canvas
// TODO: unify canvas->updates
			canvas->update ();
		} else if (id == init_timer) {
			tm->killTimer (init_timer);
			init_timer = -1;
			init ();
		}

	}
}

void Taxipilot::do_chores()
{								// meter-updates... ( speed taken care of in do_gravity () )
	if (passenger_waiting || passenger_onboard) {	// run tip as soon as a passenger is waiting
		tipLCD->setValue (tip -= tip_step);
		if (passenger_onboard) {	// run fare as soon as passenger has boarded
			fareLCD->setValue (fare);
		}
	}
	taxi->stop_flying_animation();

/* Begin Keyboard-Input-Handling-Section
	This has to be after taxi->stop_flying_animation (), so the flying_animation remains visible. */
	if (current_key_lr || current_key_ud) {

		if (passenger_onboard) {
			fare += level->fare_step + level->fare_step;
		}

		switch (current_key_lr) {	// left or right
		case -1:
			taxi->left(-ACCELERATION * acceleration);
			break;
		case 1:
			taxi->right(+ACCELERATION * acceleration);
			break;
		default:{
				if (passenger_onboard) {
					fare -= level->fare_step;
				}
			}
		}

		switch (current_key_ud) {	// up or down
		case -1:
			taxi->up(-ACCELERATION * acceleration);
			break;
		case 1:
			taxi->down(+ACCELERATION * acceleration);
			break;
		default:{
				if (passenger_onboard) {
					fare -= level->fare_step;
				}
			}
		}

		fuelDial->setValue(taxi->fuel_state());
	}

	if (toggle_gear) {
		taxi->toggle_gear();
		toggle_gear = false;
	}
/* End Keyboard-Input-Handling-Section */
}

/** Check, whether we landed on a platform and on which one */
int Taxipilot::check_landed()
{
	for (unsigned int i = 0; i < level->num_platforms; i++) {
		if (level->platform[i]->check_landed (taxi)) {
			return (i);
		}
	}
	return (-1);
}

/** Passenger has finished boarding */
void Taxipilot::picked_up()
{
	passenger_onboard = true;
	passenger_waiting = false;

	int rand_num;
	do {
		rand_num = rand() % (level->num_platforms);
	} while ((rand_num == active_platform) || (!level->platform[rand_num]->accepts_passenger));	// generate different platform
	active_platform = rand_num;

	destinationLabel->setText (level->platform[active_platform]->label);
	statusLabel->setText (i18n ("Delivery"));

	fare = level->base_fare;
}

/** Last passenger has now disappeared, a new one may be generated */
void Taxipilot::delivery_complete()
{

	fareLCD->setValue (fare = 0);	// update display
	tipLCD->setValue (tip = 0);

	passenger->discard ();
	passenger_exists = false;
	if (level->deliveries_to_go) {
		passenger_timer = tm->startTimer (this, PASSENGER_TP);	// to be randomized
	} else {					// this level is complete!
		int message_id = NEXT_LEVEL;

		if (level->extra_life) {
			message_id += EXTRA_LIFE;
			shipsLCD->display(++ships);
		}
		if ((unsigned int) (current_level+1) >= levels.count ()) {
			message_id += MISSION_COMPLETE;	// abort will be done by do_event_after_message ()
		}

		display_message (message_id);
	}
}

	/* passenger is now ready for pick-up */
void Taxipilot::passenger_ready()
{
	destinationLabel->setText(level->platform[active_platform]->label);
	statusLabel->setText(i18n("Pick up"));
	tip = level->min_base_tip + (rand() * (level->max_base_tip - level->min_base_tip) / (RAND_MAX + 1.0));
	tip_step = 0.2 / (float) cdp->cps () * passenger->impatience;

	passenger_waiting = true;
	if (onActivePlatform()) {	// in case we are already there
		passenger->pick_up (taxi->leftEdge() + (int) (taxi->width() / 2));
	}
}

  /** Computes the gravity acting on the taxi and sets it. Also updates the speedDial */
void Taxipilot::do_gravity()
{

	if (taxi->ready) {
		float xv = taxi->xVelocity() + level->horiz_gravity;	// do the basic gravity right in place;
		float yv = taxi->yVelocity() + level->base_gravity;

		int taxi_pos_x = taxi->leftEdge() + (taxi->width() / 2);
		int taxi_pos_y = taxi->topEdge() + (taxi->height() / 2);

		for (unsigned int i = 0; i < level->num_grav_centers; i++) {	// do the advanced gravity
			level->grav_center[i]->calculate_gravity (taxi_pos_x, taxi_pos_y, &xv, &yv);
		}
		
		taxi->reposition(-1, -1, xv, yv);	// this takes care of gear as well.

		speedDial->setValue((int) (16 * (taxi->speed = (sqrt(xv * xv + yv * yv)))));
	}
}

void Taxipilot::landing_func () {
	// first of all: stop
	speedDial->setValue((int) (taxi->speed = 0));
	taxi->reposition(-1, level->platform[on_platform]->current_plat_y (), 0, 0, true);
	taxi->land (level->platform[on_platform]);

	// now let's see, if there's any fuel here
	if (level->platform[on_platform]->fuel_step > 0) {
		fueling = true;			// timerEvent will take care of the actual fueling
	} else {
		fueling = false;		// fueling is only looked at while landed, so it's good enough to set it when landing
		if (taxi->fuel_out) {
			display_message(LOW_FUEL);
		}
	}

	// finally let's see, whether we are on the active platform
	if (on_platform == active_platform) {	// we landed on a meaningful platform

		if (passenger_waiting) {

			passenger->pick_up (taxi->leftEdge() + (int) (taxi->width() / 2));	// you may now come aboard!
		} else if (passenger_onboard) {

			passenger->got_delivered ((taxi->leftEdge() + (int) (taxi->width() / 2)), level->platform[active_platform]); 	// there you go!
			passenger_onboard = false;
			passenger_waiting = false;

			if ((tip + fare) >= 0) {	// we won't be substracting money
				scoreLCD->setValue (score += (tip + fare));
				if (tip >= 0) {
					deliveries_LCD->display (--(level->deliveries_to_go));	// one successful delivery
				}
			}					// TODO: else complain about being too slow

			active_platform = -1;	// delete active platform

			destinationLabel->setText(i18n("Nowhere"));	// update display
			statusLabel->setText(i18n("Idle"));
		}
	}
}

void Taxipilot::crash (int how) {
	int message_id = CRASH;

	if (--ships < 0) {
		message_id += GAME_OVER;
	}

	taxi->crash (how & CRASH_BACKGR);
	if (how & CRASH_PASSENGER) {
		passenger->die ();
	}

	display_message (message_id + how);
}

void Taxipilot::newGame () {
	delete_start_screen ();
	cdp->set_for_real (true);

	app->gameNew->setEnabled (false);
	app->gameAbort->setEnabled (true);
	app->gameSave->setEnabled (true);

	createPassengers ();

	if (cdp->saved_level < 0) {		// plain new game
		start_level = 0;
		if (!newLevel (0)) return;
		
		taxi->initialize (taxi_initial_fuel, taxi_fuel_capacity);
		
		ships = initial_ships;
	} else {			// starting a saved game
		start_level = cdp->saved_level;
		if (!newLevel (cdp->saved_level)) return;
		
		taxi->initialize (cdp->saved_fuel, taxi_fuel_capacity);
		
		ships = cdp->saved_lives;

		/* don't start same loaded game again next time */
		cdp->saved_level = -1;
	}

	fuelDial->setValue (taxi->fuel_state());
	scoreLCD->setValue (score = 0);
	shipsLCD->display (ships);

	// initialize timers and variables
	base_timer = tm->startTimer (this, 1000 / cdp->cps ());

	lastStatus = i18n("Idle");
}

/** Switches to the specified level. If index==-1, the next level is used, otherwise the level specified by index. Previous level (if any) will be deleted.
	Aborts game (+ returns false) if critical errors occured, resumeGame()s otherwise */
bool Taxipilot::newLevel (int index) {
	if (index < 0) {
		current_level++;
	} else {
		current_level = index;
	}

	useTaxi (levels[current_level].taxi_id);
	taxi->landed = false;	// otherwise taxi would try to detach itself from (to be deleted) platform

	bool splash = false;
	if (levels[current_level].splash) {
		splash = true;
	}

	deleteScreen ();

	if (splash) {
		if (levels[current_level].filename != "") {
			SplashScreen *dummy = new SplashScreen (levels[current_level].filename, true);
			dummy->setBackground (levels[current_level].splash->background_file);
			dummy->setForeground (levels[current_level].splash->foreground_file);
			dummy->setTimeout (levels[current_level].splash->timeout);
			screen = dummy;
		} else {
			screen = new SplashScreen (levels[current_level].splash->background_file, levels[current_level].splash->foreground_file, levels[current_level].splash->timeout);
			level = 0;
		}
	} else {
		screen = level = new Level(levels[current_level].filename);

		current_offset_x = level->initial_offset_x;
		current_offset_y = level->initial_offset_y;
	}

	screen->show ();
	fader->setSize(canvas->width(), canvas->height());

	if (!cdp->ok ()) {
		abortGame ();
		return false;
	}

	passenger_waiting = false;
	passenger_onboard = false;
	passenger_exists = false;
	fueling = false;

	if (level) {
		deliveries_LCD->display (level->deliveries_to_go);
	} else {
		deliveries_LCD->display (0);
	}

	game_running = true;
	resumeGame ();

	if (!splash) {
		taxi->teleport_to (level->initial_x, level->initial_y, 0, 0, true);
		passenger_timer = tm->startTimer(this, PASSENGER_TP);	// randomize ?
	} else {
		taxi->teleport_to (-100, -100, 0, 0, true);	// place taxi outside view
		event_after_message = NEXT_LEVEL;
		message_handler->displayMessage (levels[current_level].splash->message, app->ack_key_name, view->height () / 3, view->width () / 2);
		game_paused = true;
	}

	return true;
}


/** Guess what! */
void Taxipilot::quitGame () {
	pauseGame(true);
	if (3 == KMessageBox::questionYesNo(this, i18n("Really wanna quit?"), "", i18n("'course"), i18n("Never!"))) {
		kapp->quit ();
	}
	resumeGame();
}

void Taxipilot::saveGame () {
	if (current_level < 0) {
		cdp->debug_msg (i18n ("saveGame () called while no game running. This should not happen"), Serious, 0);
	}
	
	saves->addSaveGame (current_mission, mission_title, mission_version, current_level, ships, taxi->fuel, score);
	
	KMessageBox::information (0, i18n ("The game was saved successfully.\nNote that only the current level, fuel-state and number of lives are saved, not the exact game state."), i18n ("Game saved"));
}

void Taxipilot::loadGame () {
	if (game_running) {
		ask_abortGame();
	}
	if (!game_running) {		// this is NOT an else! If game is still running now, user cancelled:
		delete_start_screen ();

		MissionSelector *sel = new MissionSelector (false, MissionSelector::Saves);
		if (sel->exec () == QDialog::Accepted) {
			readMission (cdp->getMission ());
		}
		delete sel;

		initMission ();
	}
}

/** Pauses the game. */
void Taxipilot::pauseGame (bool show_pause_message) {
	if (game_running && !game_paused) {
		app->gamePause->setEnabled(false);
		if (show_pause_message) {
			display_message (PAUSE);
		}
		lastStatus = statusLabel->text();
		statusLabel->setText(i18n("Game paused"));
		fader->show();
		canvas->update();
		tm->pause ();
	}
	game_paused = true;
}

/** Resumes a paused game */
void Taxipilot::resumeGame()
{
	app->gamePause->setEnabled(true);
	canvas_timer = tm->setLastTimer (this, cdp->framePeriod ());
	statusLabel->setText(lastStatus);
	fader->hide();
	tm->resume ();
	if (game_running) {
		game_paused = false;
	}
}

/** Selects a new Mission (i.e. a list of level- and
passenger-description-files) */
void Taxipilot::selectMission()
{
	if (game_running) {
		ask_abortGame();
	}
	if (!game_running) {		// this is NOT an else! If game is still running now, user cancelled:
		delete_start_screen ();

		MissionSelector *sel = new MissionSelector ();
		if (sel->exec () == QDialog::Accepted) {
			readMission(cdp->getMission ());
		}
		delete sel;

		initMission ();
	}
}

/** Reads in a mission-description file. */
void Taxipilot::readMission(const QString & file)
{

	QDomElement e;				// just some element
	QDomNodeList list;			// just a list of nodes

	QDomElement xdocElement = cdp->open_config_file(file);
	current_mission = file;
	mission_author = cdp->get_string_attribute ("author", xdocElement, i18n ("Anonymous"), Info);
	mission_title = cdp->get_string_attribute ("title", xdocElement, i18n ("Untitled"), Info);
	mission_version = cdp->get_string_attribute ("version", xdocElement, "", Info);

	startscreen_file = cdp->get_string_attribute ("startscreen", xdocElement, bad_string_value, Serious);

	e = cdp->get_element("initials", xdocElement, Serious);	// find the initials-section
	shipsLCD->display((ships = initial_ships = cdp->get_int_attribute ("ships", e, 1, 99, 3, Serious)));

// Read passengers
	e = cdp->get_element("passengers", xdocElement, Critical);	// find the passengers-section
	list = cdp->get_node_list("passenger", e, 1, -1, 1, Critical);	// get all passenger files

	deletePassengers (); // in case previous passengers still exist

	unsigned int i = 0;
	while (i < list.length ()) {
		e = list.item (i).toElement();

		PassengerHandle dummy;
		dummy.filename = cdp->prefix_check(cdp->get_string_attribute("file", e, bad_string_value, Serious), Serious);
		QString id = cdp->get_string_attribute ("id", e, "none", Info, false);
		int x = 0;
		QString numstring = "";
		while (passengers.contains (id + numstring)) {
			if (id != "none") {
				cdp->debug_msg (i18n ("Passenger named '%1' already exists. Trying a different name. Passenger might not get selected appropriately!").arg(id), Warn, &e);
			}
			numstring.setNum (x++);
		}
		id.append (numstring);

		dummy.likelyhood = cdp->get_int_attribute ("likelyhood", e, 1, 100, 100, Info);

		if (dummy.on_start_screen = cdp->get_bool_attribute("on_start_screen", e, false, Info)) {
			dummy.x = cdp->get_int_attribute("start_screen_x", e, -100, 3000, 200, Serious);
			dummy.y = cdp->get_int_attribute("start_screen_y", e, -100, 3000, 200, Serious);
		}
		passengers.insert (id, dummy);

		i++;
	}

// Read taxis
	e = cdp->get_element ("taxis", xdocElement, Critical);
	taxi_fuel_capacity = cdp->get_float_attribute ("fuel_capacity", e, 100, 1000000, 1000, Warn);
	taxi_initial_fuel = cdp->get_float_attribute ("initial_fuel", e, 100, taxi_fuel_capacity, taxi_fuel_capacity, Warn);

	list = cdp->get_node_list ("taxi", e, 1, -1, 1, Critical);

    taxis.clear ();

	i = 0;
	while (i < list.length ()) {
		e = list.item (i).toElement();

		TaxiHandle dummy;
		dummy.filename = cdp->prefix_check(cdp->get_string_attribute("file", e, bad_string_value, Critical), Critical);
		QString taxi_id;
		taxi_id = cdp->get_string_attribute ("id", e, "none", Info, false);

		int x = 0;
		QString numstring = "";
		while (taxis.contains (taxi_id + numstring)) {
			if (taxi_id != "none") {
				cdp->debug_msg (i18n ("Taxi named '%1' already exists. Trying a different name. Taxi might not get selected appropriately!").arg(taxi_id), Warn, &e);
			}
			numstring.setNum (x++);
		}
		taxi_id.append (numstring);

		if (dummy.on_start_screen = cdp->get_bool_attribute("on_start_screen", e, false, Info)) {
			dummy.x = cdp->get_int_attribute("start_screen_x", e, -100, 3000, 200, Serious);
			dummy.y = cdp->get_int_attribute("start_screen_y", e, -100, 3000, 200, Serious);
		}
		taxis.insert (taxi_id, dummy);

		i++;
	}

// Read levels/splashes
	e = cdp->get_element("levels", xdocElement, Critical);	// find the levels-section
	list = cdp->get_node_list("", e, 1, -1, 1, Critical);	// get all level/splash files

	levels.clear ();
	splashes.clear ();

	i = 0;
	while (i < list.length ()) {
		e = list.item (i).toElement();

		LevelHandle dummy;
		dummy.splash = 0;

		if (e.tagName () == "level") {
			dummy.filename = cdp->prefix_check(cdp->get_string_attribute("file", e, bad_string_value, Serious), Serious);
		} else if (e.tagName () == "splash") {
			dummy.splash = new SplashHandle;
			DebugLevel debug_level = Serious;
			if (e.hasAttribute ("file")) {
				dummy.filename = cdp->prefix_check(cdp->get_string_attribute("file", e, bad_string_value, Serious), Serious);
				debug_level = Info;
			}
			dummy.splash->background_file = cdp->prefix_check(cdp->get_string_attribute("background", e, "", debug_level), debug_level);
			dummy.splash->foreground_file = cdp->prefix_check(cdp->get_string_attribute("foreground", e, "", Info), Info);
			dummy.splash->message = cdp->get_string_attribute("message", e, "", Info);
			dummy.splash->timeout = cdp->get_int_attribute("timeout", e, 0, 999999, 0, Info);
			splashes.append (dummy.splash);
		} else {
			cdp->debug_msg (i18n ("Unsupported tag-name '%1'. Ignoring").arg(e.tagName()), Warn, &e);
			i++;
			continue;
		}

		dummy.taxi_id = cdp->get_string_attribute ("taxi", e, "default", Info);
		if (dummy.taxi_id != "default") {
			if (!taxis.contains (dummy.taxi_id)) {
				cdp->debug_msg (i18n ("No taxi named '%1' defined. Will use default taxi instead.").arg(dummy.taxi_id), Warn, &e);
				dummy.taxi_id = "default";
			}
		}
		dummy.passenger_ids = cdp->getStringListAttribute ("passengers", e, ";", "any", Info, false);
		if (!dummy.passenger_ids.contains ("any")) {
			bool one_ok = false;
			for (unsigned int j=0; j < dummy.passenger_ids.count (); j++) {
				if (!passengers.contains (dummy.passenger_ids[j])) {
					cdp->debug_msg (i18n ("No passenger named '%1' defined. Will be ignored.").arg(dummy.passenger_ids[j]), Warn, &e);
				} else {
					one_ok = true;
				}
			}
			if (!one_ok) {
				dummy.passenger_ids.append ("any");
			}
		}
		levels.append (dummy);

		i++;
	}
}

Taxipilot::MissionDescription Taxipilot::getMissionDescription (const QString &file) {
	MissionDescription description;

	QDomElement e;				// just some element

	description.ok = false;
	QDomElement xdocElement = cdp->open_config_file(file);
	if (xdocElement.isNull ()) {
		return description;
	}

	QString version = cdp->get_string_attribute("version", xdocElement, "0.0.0", Info);
	if ((version > VERSION) || (version < "0.9.0")) {
		cdp->debug_msg (i18n ("Mission not supported by this version of Taxipilot"), Warn);
		return description;
	}
	description.splash_image = QFileInfo (file).dirPath (true) + "/" + cdp->get_string_attribute ("startscreen", xdocElement, bad_string_value, Serious);

	description.ok = true;		// mission could at least be opened

	e = cdp->get_element("abstract", xdocElement, Warn);
	description.abstract = cdp->getElementText (e, i18n ("No description available for this mission"), Warn);

	description.author = cdp->get_string_attribute("author", xdocElement, i18n("Anonymous"), Info);
	description.title = cdp->get_string_attribute("title", xdocElement, i18n("Untitled"), Info);

	return description;
}

/** Asks whether game should really be aborted, then calls abortGame () */
void Taxipilot::ask_abortGame()
{

	pauseGame(true);
	if (3 != KMessageBox::questionYesNo(this, i18n("This will quit the current game!\nAre you sure?"))) {
//		resumeGame();
		return;
	}
	abortGame();
}

/** This function aborts the current game. Usually called by ask_abortGame, which
first asks, then calls this function */
void Taxipilot::abortGame()
{
	pauseGame(false);				// Usually we've already done this in ask_abortGame. It should'nt harm, though

	message_handler->kill ();
	game_running = false;
	tm->killTimers(this);

	if (passenger_exists) {
		passenger->discard();
	}

	deleteScreen ();
	deleteTaxi ();

	app->gamePause->setEnabled (false);
	app->gameAbort->setEnabled (false);
	app->gameSave->setEnabled (false);
	app->gameNew->setEnabled (true);

// in case there is still a message
	fader->hide();
	canvas->update();

	base_timer = -1;
	passenger_timer = -1;
	canvas_timer = -1;
	current_level = -1;

	initMission ();
}

/** Creates all passengers */
void Taxipilot::createPassengers () {
	if (!passengers_created) {                         // create all passengers
		Passengers::Iterator it;
		for (it = passengers.begin (); it != passengers.end (); it++) {
			it.data ().pointer = new Passenger (it.data ().filename);
		}
		passengers_created = true;
	}
}

/** Deletes all passengers (if they were created) */
void Taxipilot::deletePassengers () {
	if (passengers_created) {
		Passengers::Iterator it;
		for (it = passengers.begin (); it != passengers.end (); it++) {
			delete it.data().pointer;
		}
		passengers_created = false;
	}
   	passengers.clear ();
}

/** Selects a new passenger based on the provided category */
Passenger *Taxipilot::selectPassenger (const QStringList &ids) {
	int number, chance;

	QStringList use_ids = ids;
	if (ids.contains ("any")) {
		use_ids = passengers.keys ();
	}

	while (true) {
		number = (rand() % use_ids.count ());
		chance = (rand() % 100);
		if (chance < passengers[use_ids[number]].likelyhood) {
			return passengers[use_ids[number]].pointer;
		}
	}
}

/** Switches to the taxi indicated by id */
void Taxipilot::useTaxi (const QString &id) {
	QString use_id = id;
	if (!taxis.contains (id)) {
		use_id = taxis.begin ().key ();
	}
	if (use_id != current_taxi) {
		Taxi *new_taxi = new Taxi (taxis[use_id].filename);
		if (taxi) {
			new_taxi->initialize (taxi->fuel, taxi_fuel_capacity);
		}
		deleteTaxi ();
		taxi = new_taxi;
		current_taxi = use_id;
	}
}

/** deletes the current taxi. Safe wrapper around "delete taxi" */
void Taxipilot::deleteTaxi () {
	if (taxi) {
		delete taxi;
		taxi = 0;
	}
	current_taxi = "";
}

/** deletes the current screen. Safe wrapper around "delete screen" */
void Taxipilot::deleteScreen () {
	if (screen) {
		delete screen;
		screen = 0;
	}
}

/** Checks, whether everything is ok, en/disables Game->New accordingly and displays
	startscreen */
void Taxipilot::initMission () {
	QString status = i18n("Error");

	cdp->setCurrentFile (current_mission);

	if (cdp->ok()) {
		display_start_screen();
		if (cdp->ok ()) {				// things can go wrong while creating the start-screen!
			event_after_message = NEW_GAME;
			QString dummy = i18n ("To start game:");
			message_handler->displayMessage (dummy, app->ack_key_name, view->height () / 3, view->width () / 3);
			app->gameNew->setEnabled(true);
			status = i18n("No Game running");
			cdp->setLastLoaded (current_mission);
		} else {
			app->gameNew->setEnabled(false);
			pauseGame (false);
			delete_start_screen();
		}
	} else {
		app->gameNew->setEnabled(false);
	}

	statusLabel->setText (status);
	
	if (cdp->saved_level >= 0) {		// a saved game was loaded. let's start it right away
		newGame ();
		pauseGame ();
	}
}


/** Generates and displays the current mission's startscreen */
void Taxipilot::display_start_screen()
{
	cdp->set_for_real (false);

	if (start_screen_active)
		return;
	start_screen_active = true;

	start_screen = cdp->createQCPixmap (startscreen_file, 640, 480, Serious);

	canvas->setBackgroundPixmap(*start_screen);
	canvas->resize(start_screen->width(), start_screen->height());

	view->setFixedSize(canvas->width() + 3, canvas->height() + 3);

	Passengers::Iterator it;
	for (it = passengers.begin (); it != passengers.end (); it++) {
		if (it.data ().on_start_screen) {
			Passenger *p = new Passenger (it.data ().filename);
			p->initialize (it.data ().x, it.data ().y, 0);
			passengers_on_start_screen.append (p);
		}
	}

	Taxis::Iterator t_it;
	for (t_it = taxis.begin (); t_it != taxis.end (); t_it++) {
		if (t_it.data ().on_start_screen) {
			Taxi *t = new Taxi (t_it.data ().filename, t_it.data ().x, t_it.data ().y);
			taxis_on_start_screen.append (t);
		}
	}

	resumeGame ();		// restart timers

	app->setCaption(mission_title);

	credits = new Credits (mission_title, mission_author, taxis, passengers, levels);
}

/** Delete (the passengers on) the Startscreen */
void Taxipilot::delete_start_screen()
{
	if (!start_screen_active)
		return;
	start_screen_active = false;

	passengers_on_start_screen.clear ();
	taxis_on_start_screen.clear ();

	delete credits;

	canvas->setBackgroundColor (QColor (0, 0, 0));
	delete start_screen;
	message_handler->kill ();
}

/** Displays a message as QCanvasText on the Canvas (scrolling it into view). */
void Taxipilot::display_message(int message_id)
{

	QString message = "";

	if (message_id != PAUSE) {
		pauseGame(false);
	} else {
		message = i18n ("Game paused\nto continue:");
	}

	event_after_message = message_id;

	if (message_id & CRASH) {
		if (message_id & CRASH_PASSENGER) {
			message = i18n("You just killed an innocent passenger!\nAre you mad?!?");
		} else if (message_id & CRASH_BACKGR) {
			message = i18n("You just crashed your taxi.\nThat's gonna be expensive!");
		}
		if (message_id & GAME_OVER) {
			message += i18n("\n\nYou crashed so many times now...\n...go find another job!\nGame Over!");
		} else if (passenger_onboard) {
			message += i18n("\n\nYour passenger decided to walk the rest of the way.");
		}
		if (taxi->fuel_out) {
			message += i18n("\nBetter watch your fuel next time.");
		}
	}

	if (message_id & NEXT_LEVEL) {
		message = i18n("Congratulations on finishing this level!\n");
		if (message_id & EXTRA_LIFE) {
			message += (i18n("Additionally you are awarded a new ship.\n"));
		}
		if (message_id & MISSION_COMPLETE) {
			message += (i18n("\nYou completed your mission.\nGo do something useful now."));
		}
	}
	// Don't display fuel-message if game over, or you completed a level.
	if (!(message_id & (NEXT_LEVEL | GAME_OVER | PAUSE)) && taxi->fuel_low) {
		message += i18n("\nSince you ran low on fuel, the people at '%1'\n"
				"generously offer you some fuel at only twice the regular price.").arg(level->platform[level->exp_fuel_platform]->label);
		event_after_message |= LOW_FUEL;
	}

	message_handler->displayMessage (message, app->ack_key_name);
}

/** I really hate doing this, but I need to have a function that is called after a text
message was displayed, which will do, whatever needs to be done next.
Unfortunately, what needs to be done may vary. */
void Taxipilot::do_event_after_message()
{
	current_key_lr = current_key_ud = 0;	// in case we missed a key-release somewhere

	if (event_after_message & LOW_FUEL) {
		QString dummy;
		scoreLCD->setValue (score -= taxi->fuel_up (FUEL_UP_EMPTY_BY * cdp->cps () / 20) * level->platform[level->exp_fuel_platform]->fuel_price * 2);	// don't check whether affordable
		fuelDial->setValue(taxi->fuel_state());
	}

	if (event_after_message & (GAME_OVER | MISSION_COMPLETE)) {	// no more lives/levels
		if (event_after_message & MISSION_COMPLETE) {
			highscores->addScore (current_mission, start_level, current_level + 1, score);
		} else {
			highscores->addScore (current_mission, start_level, current_level, score);
		}
		abortGame();
		return;

	} else if (event_after_message & CRASH) {	// continue after crash

		taxi->teleport_to(level->initial_x, level->initial_y, 0, 0, true);
		shipsLCD->display(ships);

		if (passenger_onboard || (event_after_message & CRASH_PASSENGER)) {
			passenger_onboard = false;
			passenger_waiting = false;
			destinationLabel->setText(i18n("Nowhere"));	// update display
			lastStatus = i18n("Idle");

			if (!level->deliveries_to_go) {	// If this was the last passenger to deliver, we would get stuck.
				deliveries_LCD->display (level->deliveries_to_go = 1);
			}

			delivery_complete();	// of course it is not, but this will take care of the rest
		}

	} else if (event_after_message & NEXT_LEVEL) {	// next level
		newLevel ();
		return;		// newLevel already included resumeGame () (if successful)
	} else if (event_after_message & NEW_GAME) {
		newGame ();
		return;		// newGame already included resumeGame () (if successful)
	}

	resumeGame();				// this will set the AdvancePeriod.
}

/** Takes care of any scrolling necessary (for oversized levels) */
void Taxipilot::doScroll()
{
	int new_offset_x = current_offset_x;
	int new_offset_y = current_offset_y;

// scroll left or right
	if (taxi->leftEdge () < (current_offset_x + level->margin)) {
		new_offset_x = taxi->leftEdge () - level->margin;
		if (taxi->xVelocity () < 0.0) {
			new_offset_x += static_cast<int> (cdp->scrollStep () * taxi->xVelocity ());
		}
		if (new_offset_x < 0) new_offset_x = 0;
	} else if (taxi->rightEdge () > (current_offset_x + level->visible_width - level->margin)) {
		new_offset_x = taxi->rightEdge () - level->visible_width + level->margin;
		if (taxi->xVelocity () > 0.0) {
			new_offset_x += static_cast<int> (cdp->scrollStep () * taxi->xVelocity ());
		}
		if (new_offset_x > level->max_offset_x) new_offset_x = level->max_offset_x;
	}

// scroll up or down
	if (taxi->topEdge () < (current_offset_y + level->margin)) {
		new_offset_y = taxi->topEdge () - level->margin;
		if (taxi->yVelocity () < 0.0) {
			new_offset_y += static_cast<int> (cdp->scrollStep () * taxi->yVelocity ());
		}
		if (new_offset_y < 0) new_offset_y = 0;
	} else if (taxi->bottomEdge () > (current_offset_y + level->visible_height - level->margin)) {
		new_offset_y = taxi->bottomEdge () - level->visible_height + level->margin;
		if (taxi->yVelocity () > 0.0) {
			new_offset_y += static_cast<int> (cdp->scrollStep () * taxi->yVelocity ());
		}
		if (new_offset_y > level->max_offset_y) new_offset_y = level->max_offset_y;
	}

	if (!((new_offset_x == current_offset_x) && (new_offset_y == current_offset_y))) {
		current_offset_x = new_offset_x;
		current_offset_y = new_offset_y;

		view->setContentsPos(current_offset_x, current_offset_y);
	}
}

/** Pause the game when focus gets lost. */
void Taxipilot::focusOutEvent (QFocusEvent *){
	if (!game_paused) app->gamePause->activate();
}

void Taxipilot::configChanged(){
	sw->init (cdp->sound_enabled ());

	float x = (float) cdp->fps () * (float) cdp->cps ();
	acceleration = 2000 / x;
	if (passenger_exists) {
		tip_step = 0.2 / (float) cdp->cps () * passenger->impatience;
	}
	if (game_running) {
		level->config_changed ();	// the level will tell the centers of gravity
		taxi->config_changed ();
	}
	if (base_timer != -1) {
		tm->killTimer (base_timer);
	}
	base_timer = tm->startTimer (this, 1000 / cdp->cps ());		// AdvancePeriod is set in resumeGame ()

	speedDial->setMaxValue (10000 / cdp->fps ());
	speedDial->setLimits (-1, 4000 / cdp->fps ());
	speedDial->setLineStep (1000 / cdp->fps ());
	speedDial->setPageStep (2000 / cdp->fps ());
	speedDial->setNotchTarget (5);

	// canvas-timer will be set in resumeGame ()
}

void Taxipilot::screenDone () {
	do_event_after_message ();	// event_after_message should be NEXT_LEVEL
	// TODO: rewrite entire code for triggering of next level.
}

void Taxipilot::showHighscores () {
	highscores->showHighscores (current_mission);
}

#include "taxipilot.moc"
